"""
symbolite.abstract
~~~~~~~~~~~~~~~~~~

Abstract symbolite primitives.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .scalar import Scalar
from .symbol import Function, Symbol, UserFunction
from .vector import Vector

__all__ = ["Symbol", "Function", "Scalar", "Vector", "UserFunction"]



"""
symbolite.abstract.scalar
~~~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for scalar operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import dataclasses
from typing import Any

from ..core import Unsupported
from . import symbol
from .symbol import BaseFunction, Symbol, downcast

NumberT = int | float | complex


@dataclasses.dataclass(frozen=True, repr=False)
class Scalar(Symbol):
    """A user defined symbol."""

    def __getitem__(self, key: Any):
        return Unsupported

    def __getattr__(self, key: Any):
        raise AttributeError(key)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Scalar:
        """Implements addition."""
        return downcast(symbol.add(self, other), Scalar)

    def __sub__(self, other: Any) -> Scalar:
        """Implements subtraction."""
        return downcast(symbol.sub(self, other), Scalar)

    def __mul__(self, other: Any) -> Scalar:
        """Implements multiplication."""
        return downcast(symbol.mul(self, other), Scalar)

    def __matmul__(self, other: Any) -> Scalar:
        """Implements multiplication."""
        return downcast(symbol.matmul(self, other), Scalar)

    def __truediv__(self, other: Any) -> Scalar:
        """Implements true division."""
        return downcast(symbol.truediv(self, other), Scalar)

    def __floordiv__(self, other: Any) -> Scalar:
        """Implements integer division using the // operator."""
        return downcast(symbol.floordiv(self, other), Scalar)

    def __mod__(self, other: Any) -> Scalar:
        """Implements modulo using the % operator."""
        return downcast(symbol.mod(self, other), Scalar)

    def __pow__(self, other: Any, modulo: Any = None) -> Scalar:
        """Implements behavior for exponents using the ** operator."""
        if modulo is None:
            return downcast(symbol.pow(self, other), Scalar)
        else:
            return downcast(symbol.pow3(self, other, modulo), Scalar)

    def __lshift__(self, other: Any) -> Scalar:
        """Implements left bitwise shift using the << operator."""
        return downcast(symbol.lshift(self, other), Scalar)

    def __rshift__(self, other: Any) -> Scalar:
        """Implements right bitwise shift using the >> operator."""
        return downcast(symbol.rshift(self, other), Scalar)

    def __and__(self, other: Any) -> Scalar:
        """Implements bitwise and using the & operator."""
        return downcast(symbol.and_(self, other), Scalar)

    def __or__(self, other: Any) -> Scalar:
        """Implements bitwise or using the | operator."""
        return downcast(symbol.or_(self, other), Scalar)

    def __xor__(self, other: Any) -> Scalar:
        """Implements bitwise xor using the ^ operator."""
        return downcast(symbol.xor(self, other), Scalar)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Scalar:
        """Implements reflected addition."""
        return downcast(symbol.add(other, self), Scalar)

    def __rsub__(self, other: Any) -> Scalar:
        """Implements reflected subtraction."""
        return downcast(symbol.sub(other, self), Scalar)

    def __rmul__(self, other: Any) -> Scalar:
        """Implements reflected multiplication."""
        return downcast(symbol.mul(other, self), Scalar)

    def __rmatmul__(self, other: Any) -> Scalar:
        """Implements reflected multiplication."""
        return downcast(symbol.matmul(other, self), Scalar)

    def __rtruediv__(self, other: Any) -> Scalar:
        """Implements reflected true division."""
        return downcast(symbol.truediv(other, self), Scalar)

    def __rfloordiv__(self, other: Any) -> Scalar:
        """Implements reflected integer division using the // operator."""
        return downcast(symbol.floordiv(other, self), Scalar)

    def __rmod__(self, other: Any) -> Scalar:
        """Implements reflected modulo using the % operator."""
        return downcast(symbol.mod(other, self), Scalar)

    def __rpow__(self, other: Any) -> Scalar:
        """Implements behavior for reflected exponents using the ** operator."""
        return downcast(symbol.pow(other, self), Scalar)

    def __rlshift__(self, other: Any) -> Scalar:
        """Implements reflected left bitwise shift using the << operator."""
        return downcast(symbol.lshift(other, self), Scalar)

    def __rrshift__(self, other: Any) -> Scalar:
        """Implements reflected right bitwise shift using the >> operator."""
        return downcast(symbol.rshift(other, self), Scalar)

    def __rand__(self, other: Any) -> Scalar:
        """Implements reflected bitwise and using the & operator."""
        return downcast(symbol.and_(other, self), Scalar)

    def __ror__(self, other: Any) -> Scalar:
        """Implements reflected bitwise or using the | operator."""
        return downcast(symbol.or_(other, self), Scalar)

    def __rxor__(self, other: Any) -> Scalar:
        """Implements reflected bitwise xor using the ^ operator."""
        return downcast(symbol.xor(other, self), Scalar)

    # Unary operators and functions
    def __neg__(self) -> Scalar:
        """Implements behavior for negation (e.g. -some_object)"""
        return downcast(symbol.neg(self), Scalar)

    def __pos__(self) -> Scalar:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return downcast(symbol.pos(self), Scalar)

    def __invert__(self) -> Scalar:
        """Implements behavior for inversion using the ~ operator."""
        return downcast(symbol.invert(self), Scalar)


@dataclasses.dataclass(frozen=True, repr=False)
class ScalarUnaryFunction(BaseFunction):
    namespace: str = "scalar"
    arity: int = 1

    def __call__(self, arg1: Scalar | NumberT) -> Scalar:
        return super()._call(arg1)  # type: ignore

    @property
    def output_type(self):
        return Scalar


@dataclasses.dataclass(frozen=True, repr=False)
class ScalarBinaryFunction(BaseFunction):
    namespace: str = "scalar"
    arity: int = 2

    def __call__(self, arg1: Scalar | NumberT, arg2: Scalar | NumberT) -> Scalar:
        return super()._call(arg1, arg2)  # type: ignore

    @property
    def output_type(self):
        return Scalar


# "gcd": None,  # 1 to ---
# "hypot": None,  # 1 to ---
# "isclose": None,  # 2, 3, 4
# "lcm": None,  # 1 to ---
# "perm": None,  # 1 or 2
# "log": None,  # 1 or 2 is used as log(x, e)

abs = ScalarUnaryFunction("abs")
acos = ScalarUnaryFunction("acos")
acosh = ScalarUnaryFunction("acosh")
asin = ScalarUnaryFunction("asin")
asinh = ScalarUnaryFunction("asinh")
atan = ScalarUnaryFunction("atan")
atan2 = ScalarBinaryFunction("atan2")
atanh = ScalarUnaryFunction("atanh")
ceil = ScalarUnaryFunction("ceil")
comb = ScalarUnaryFunction("comb")
copysign = ScalarUnaryFunction("copysign")
cos = ScalarUnaryFunction("cos")
cosh = ScalarUnaryFunction("cosh")
degrees = ScalarUnaryFunction("degrees")
erf = ScalarUnaryFunction("erf")
erfc = ScalarUnaryFunction("erfc")
exp = ScalarUnaryFunction("exp")
expm1 = ScalarUnaryFunction("expm1")
fabs = ScalarUnaryFunction("fabs")
factorial = ScalarUnaryFunction("factorial")
floor = ScalarUnaryFunction("floor")
fmod = ScalarUnaryFunction("fmod")
frexp = ScalarUnaryFunction("frexp")
gamma = ScalarUnaryFunction("gamma")
hypot = ScalarUnaryFunction("gamma")
isfinite = ScalarUnaryFunction("isfinite")
isinf = ScalarUnaryFunction("isinf")
isnan = ScalarUnaryFunction("isnan")
isqrt = ScalarUnaryFunction("isqrt")
ldexp = ScalarBinaryFunction("ldexp")
lgamma = ScalarUnaryFunction("lgamma")
log = ScalarUnaryFunction("log")
log10 = ScalarUnaryFunction("log10")
log1p = ScalarUnaryFunction("log1p")
log2 = ScalarUnaryFunction("log2")
modf = ScalarUnaryFunction("modf")
nextafter = ScalarUnaryFunction("nextafter")
pow = ScalarUnaryFunction("pow")
radians = ScalarUnaryFunction("radians")
remainder = ScalarBinaryFunction("remainder")
sin = ScalarUnaryFunction("sin")
sinh = ScalarUnaryFunction("sinh")
sqrt = ScalarUnaryFunction("sqrt")
tan = ScalarUnaryFunction("tan")
tanh = ScalarUnaryFunction("tanh")
tan = ScalarUnaryFunction("tan")
trunc = ScalarUnaryFunction("trunc")
ulp = ScalarUnaryFunction("ulp")

e = Scalar("e", namespace="scalar")
inf = Scalar("inf", namespace="scalar")
pi = Scalar("pi", namespace="scalar")
nan = Scalar("nan", namespace="scalar")
tau = Scalar("tau", namespace="scalar")

del BaseFunction, Symbol, dataclasses



"""
symbolite.abstract.vector
~~~~~~~~~~~~~~~~~~~~~~~~~

Objects and functions for vector operations.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

import dataclasses
from typing import Any, Iterable, Mapping, Sequence, overload

from ..core import substitute
from . import symbol
from .scalar import NumberT, Scalar
from .symbol import BaseFunction, Symbol, downcast

VectorT = Iterable[NumberT]


@dataclasses.dataclass(frozen=True, repr=False)
class Vector(Symbol):
    """A user defined symbol."""

    def __getitem__(self, key: int | Scalar) -> Scalar:
        return downcast(super().__getitem__(key), Scalar)

    def __getattr__(self, key: Any):
        raise AttributeError(key)

    # Normal arithmetic operators
    def __add__(self, other: Any) -> Vector:
        """Implements addition."""
        return downcast(symbol.add(self, other), Vector)

    def __sub__(self, other: Any) -> Vector:
        """Implements subtraction."""
        return downcast(symbol.sub(self, other), Vector)

    def __mul__(self, other: Any) -> Vector:
        """Implements multiplication."""
        return downcast(symbol.mul(self, other), Vector)

    def __matmul__(self, other: Vector) -> Scalar:
        """Implements multiplication."""
        return downcast(symbol.matmul(self, other), Scalar)

    def __truediv__(self, other: Vector) -> Vector:
        """Implements true division."""
        return downcast(symbol.truediv(self, other), Vector)

    def __floordiv__(self, other: Any) -> Vector:
        """Implements integer division using the // operator."""
        return downcast(symbol.floordiv(self, other), Vector)

    def __mod__(self, other: Any) -> Vector:
        """Implements modulo using the % operator."""
        return downcast(symbol.mod(self, other), Vector)

    def __pow__(self, other: Any, modulo: Any = None) -> Vector:
        """Implements behavior for exponents using the ** operator."""
        if modulo is None:
            return downcast(symbol.pow(self, other), Vector)
        else:
            return downcast(symbol.pow3(self, other, modulo), Vector)

    def __lshift__(self, other: Any) -> Vector:
        """Implements left bitwise shift using the << operator."""
        return downcast(symbol.lshift(self, other), Vector)

    def __rshift__(self, other: Any) -> Vector:
        """Implements right bitwise shift using the >> operator."""
        return downcast(symbol.rshift(self, other), Vector)

    def __and__(self, other: Any) -> Vector:
        """Implements bitwise and using the & operator."""
        return downcast(symbol.and_(self, other), Vector)

    def __or__(self, other: Any) -> Vector:
        """Implements bitwise or using the | operator."""
        return downcast(symbol.or_(self, other), Vector)

    def __xor__(self, other: Any) -> Vector:
        """Implements bitwise xor using the ^ operator."""
        return downcast(symbol.xor(self, other), Vector)

    # Reflected arithmetic operators
    def __radd__(self, other: Any) -> Vector:
        """Implements reflected addition."""
        return downcast(symbol.add(other, self), Vector)

    def __rsub__(self, other: Any) -> Vector:
        """Implements reflected subtraction."""
        return downcast(symbol.sub(other, self), Vector)

    def __rmul__(self, other: Any) -> Vector:
        """Implements reflected multiplication."""
        return downcast(symbol.mul(other, self), Vector)

    def __rmatmul__(self, other: Any) -> Scalar:
        """Implements reflected multiplication."""
        return downcast(symbol.matmul(other, self), Scalar)

    def __rtruediv__(self, other: Any) -> Vector:
        """Implements reflected true division."""
        return downcast(symbol.truediv(other, self), Vector)

    def __rfloordiv__(self, other: Any) -> Vector:
        """Implements reflected integer division using the // operator."""
        return downcast(symbol.floordiv(other, self), Vector)

    def __rmod__(self, other: Any) -> Vector:
        """Implements reflected modulo using the % operator."""
        return downcast(symbol.mod(other, self), Vector)

    def __rpow__(self, other: Any) -> Vector:
        """Implements behavior for reflected exponents using the ** operator."""
        return downcast(symbol.pow(other, self), Vector)

    def __rlshift__(self, other: Any) -> Vector:
        """Implements reflected left bitwise shift using the << operator."""
        return downcast(symbol.lshift(other, self), Vector)

    def __rrshift__(self, other: Any) -> Vector:
        """Implements reflected right bitwise shift using the >> operator."""
        return downcast(symbol.rshift(other, self), Vector)

    def __rand__(self, other: Any) -> Vector:
        """Implements reflected bitwise and using the & operator."""
        return downcast(symbol.and_(other, self), Vector)

    def __ror__(self, other: Any) -> Vector:
        """Implements reflected bitwise or using the | operator."""
        return downcast(symbol.or_(other, self), Vector)

    def __rxor__(self, other: Any) -> Vector:
        """Implements reflected bitwise xor using the ^ operator."""
        return downcast(symbol.xor(other, self), Vector)

    # Unary operators and functions
    def __neg__(self) -> Vector:
        """Implements behavior for negation (e.g. -some_object)"""
        return downcast(symbol.neg(self), Vector)

    def __pos__(self) -> Vector:
        """Implements behavior for unary positive (e.g. +some_object)"""
        return downcast(symbol.pos(self), Vector)

    def __invert__(self) -> Vector:
        """Implements behavior for inversion using the ~ operator."""
        return downcast(symbol.invert(self), Vector)


@dataclasses.dataclass(frozen=True, repr=False)
class CumulativeFunction(BaseFunction):
    namespace: str = "vector"
    arity: int = 1

    def __call__(self, arg1: Vector | VectorT) -> Scalar:
        return super()._call(arg1)  # type: ignore

    @property
    def output_type(self):
        return Scalar


sum = CumulativeFunction("sum", namespace="vector")
prod = CumulativeFunction("prod", namespace="vector")


@overload
def vectorize(
    expr: NumberT,
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> NumberT: ...


@overload
def vectorize(
    expr: Symbol,
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> Symbol: ...


@overload
def vectorize(
    expr: Iterable[NumberT | Symbol],
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[NumberT | Symbol, ...]: ...


def vectorize(
    expr: NumberT | Symbol | Iterable[NumberT | Symbol],
    symbol_names: Sequence[str] | Mapping[str, int],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> NumberT | Symbol | tuple[NumberT | Symbol, ...]:
    """Vectorize expression by replacing scalar symbols
    by an array at a given indices.

    Parameters
    ----------
    expr
    symbol_names
        if a tuple, provides the names of the symbols
        which will be mapped to the indices given by their position.
        if a dict, maps symbol names to indices.
    varname
        name of the array variable
    """
    if isinstance(expr, NumberT):
        return expr

    if not isinstance(expr, Symbol):
        return tuple(vectorize(symbol, symbol_names, varname) for symbol in expr)

    if isinstance(symbol_names, dict):
        it = zip(symbol_names.values(), symbol_names.keys())
    else:
        it = enumerate(symbol_names)

    arr = Vector(varname)

    reps = {scalar_type(name): arr[ndx] for ndx, name in it}
    return substitute(expr, reps)


@overload
def auto_vectorize(
    expr: NumberT,
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], Symbol]: ...


@overload
def auto_vectorize(
    expr: Symbol,
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], Symbol]: ...


@overload
def auto_vectorize(
    expr: Iterable[Symbol],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], tuple[Symbol, ...]]: ...


def auto_vectorize(
    expr: NumberT | Symbol | Iterable[Symbol],
    varname: str = "vec",
    scalar_type: type[Scalar] = Scalar,
) -> tuple[tuple[str, ...], NumberT | Symbol | tuple[NumberT | Symbol, ...]]:
    """Vectorize expression by replacing all test_scalar symbols
    by an array at a given indices. Symbols are ordered into
    the array alphabetically.

    Parameters
    ----------
    expr
    varname
        name of the array variable

    Returns
    -------
    tuple[str, ...]
        symbol names as ordered in the array.
    SymbolicExpression
        vectorized expression.
    """
    if isinstance(expr, NumberT):
        return tuple(), expr

    if not isinstance(expr, Symbol):
        expr = tuple(expr)
        out = set[str]()
        for symbol in expr:
            out.update(symbol.symbol_names(""))
        symbol_names = tuple(sorted(out))
        return symbol_names, vectorize(expr, symbol_names, varname, scalar_type)
    else:
        symbol_names = tuple(sorted(expr.symbol_names("")))
        return symbol_names, vectorize(expr, symbol_names, varname, scalar_type)



"""
symbolite
~~~~~~~~~

A minimal symbolic python package.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from .abstract import Function, Scalar, Symbol, UserFunction, Vector, scalar, vector

__all__ = ["Symbol", "Function", "Scalar", "scalar", "Vector", "vector", "UserFunction"]



"""
symbolite.core
~~~~~~~~~~~~~~

Symbolite core classes and functions.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import collections
import types
import warnings
from functools import singledispatch
from operator import attrgetter
from typing import Any, Callable, Mapping, Sequence

from ..impl import find_module_in_stack


class Unsupported(ValueError):
    """Label unsupported"""


def as_string(expr: Any) -> str:
    """Return the expression as a string.

    Parameters
    ----------
    expr
        symbolic expression.
    """
    return str(expr)


def as_function(
    expr: Any,
    function_name: str,
    params: Sequence[str],
    libsl: types.ModuleType | None = None,
) -> Callable[..., Any]:
    """Converts the expression to a callable function.

    Parameters
    ----------
    expr
        symbolic expression.
    function_name
        name of the function to be used.
    params
        names of the parameters.
    libsl
        implementation module.
    """

    function_def = (
        f"""def {function_name}({", ".join(params)}): return {as_string(expr)}"""
    )

    lm = compile(function_def, libsl)

    return lm[function_name]


def compile(
    code: str,
    libsl: types.ModuleType | None = None,
) -> dict[str, Any]:
    """Compile the code and return the local dictionary.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from ..impl import libstd as libsl

    assert libsl is not None

    lm: dict[str, Any] = {}
    exec(
        code,
        {
            "symbol": libsl.symbol,
            "scalar": libsl.scalar,
            "vector": libsl.vector,
            **globals(),
        },
        lm,
    )
    return lm


def inspect(expr: Any) -> dict[Any, int]:
    """Inspect an expression and return what is there
    and how many times.

    Parameters
    ----------
    expr
        symbolic expression.
    """
    if hasattr(expr, "yield_named"):
        cnt = collections.Counter[Any](expr.yield_named())
        return dict(cnt)
    return {expr: 1}


@singledispatch
def evaluate_impl(expr: Any, libsl: types.ModuleType) -> Any:
    """Evaluate expression.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """
    return expr


def evaluate(expr: Any, libsl: types.ModuleType | None = None) -> Any:
    """Evaluate expression.

    Parameters
    ----------
    expr
        symbolic expression.
    libsl
        implementation module.
    """

    if libsl is None:
        libsl = find_module_in_stack()
    if libsl is None:
        warnings.warn("No libsl provided, defaulting to Python standard library.")
        from ..impl import libstd as libsl

    return evaluate_impl(expr, libsl)


@singledispatch
def substitute(expr: Any, replacements: Mapping[Any, Any]) -> Any:
    """Replace symbols, functions, values, etc by others.

    Parameters
    ----------
    expr
        symbolic expression.
    replacements
        replacement dictionary.
    """
    return replacements.get(expr, expr)


@singledispatch
def substitute_by_name(expr: Any, **replacements: Any) -> Any:
    """Replace Symbols by values or objects, matching by name.

    Parameters
    ----------
    expr
        symbolic expression.
    replacements
        replacement dictionary.
    """

    if hasattr(expr, "subs_by_name"):
        return expr.subs_by_name(**replacements)
    return expr


@evaluate_impl.register
def evaluate_impl_str(expr: str, libsl: types.ModuleType) -> Any | Unsupported:
    return attrgetter(expr)(libsl)



"""
symbolite.impl.libjax.vector
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite.abstract.vector
into values and functions defined in JAX.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

import operator

import jax.numpy as np

from symbolite.core import Unsupported

op_getitem = operator.getitem

sum = np.sum
prod = np.prod

Vector = Unsupported


del np, operator, Unsupported



"""
symbolite.impl.libnumpy
~~~~~~~~~~~~~~~~~~~~~~~

Translate symbolite
into values and functions defined in NumPy.

:copyright: 2023 by Symbolite Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from . import scalar, symbol, vector

__all__ = ["symbol", "scalar", "vector"]
